/*
 * Created on Oct 10, 2011
 *
 */
package org.gk.pathwaylayout;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.gk.graphEditor.PathwayEditor;
import org.gk.model.GKInstance;
import org.gk.model.InstanceUtilities;
import org.gk.model.PersistenceAdaptor;
import org.gk.model.ReactomeJavaConstants;
import org.gk.render.FlowLine;
import org.gk.render.HyperEdge;
import org.gk.render.Node;
import org.gk.render.Note;
import org.gk.render.ProcessNode;
import org.gk.render.Renderable;
import org.gk.render.RenderableCompartment;
import org.gk.render.RenderableComplex;
import org.gk.render.RenderablePathway;

/**
 * This customized PathwayEditor is used to draw disease related pathway.
 * @author gwu
 *
 */
public class DiseasePathwayImageEditor extends PathwayEditor {
    private boolean isForNormal;
    private PersistenceAdaptor adaptor;
    
    public DiseasePathwayImageEditor() {
    }
    
    /**
     * Set a flag to indicate that the paint method should draw objects that are related to normal 
     * pathway part only.
     * @param isForNormal
     */
    public void setIsForNormal(boolean isForNormal) {
        this.isForNormal = isForNormal;
    }
    
    /**
     * To be used to query GKInstances based on reactomeIds.
     * @param adaptor
     */
    public void setPersistenceAdaptor(PersistenceAdaptor adaptor) {
        this.adaptor = adaptor;
    }

    @Override
    @SuppressWarnings({"unchecked", "rawtypes"})
    public void paint(Graphics g) {
        Set<Long> normalIds = new HashSet<Long>();
        Set<Long> diseaseIds = new HashSet<Long>();
        splitObjects(normalIds, diseaseIds);
        
        //Clear the editor
        Graphics2D g2 = (Graphics2D)g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);
        
        // In case g is provided by other graphics context (e.g during exporting)
        g2.setFont(getFont());
        g2.setBackground(background);
        Dimension size = getPreferredSize();
        g2.clearRect(0, 0, size.width, size.height);
        // TODO: Delete the follow two statements. For test only
//        g2.setPaint(Color.yellow);
//        g2.fillRect(0, 0, size.width, size.height);
        // Enable zooming
        g2.scale(scaleX, scaleY);
        
        List comps = displayedObject.getComponents();
        // Draw nodes first
        if (comps == null || comps.size() == 0)
            return;
        java.util.List edges = new ArrayList();
        Rectangle clip = g.getClipBounds();
        // Draw Compartments first
        List<RenderableCompartment> compartments = drawCompartments(g, 
                                                                    comps, 
                                                                    null,
                                                                    clip);
        
        // Draw normal objects
        List<Renderable> diseaseComps = new ArrayList<Renderable>();
        List<Renderable> list = new ArrayList<Renderable>();
        // We need to check edges first since states of nodes depend on edges 
        for (Iterator<?> it = comps.iterator(); it.hasNext();) {
            Renderable r = (Renderable) it.next();
            if (r instanceof HyperEdge) {
                if (isNormalObject(r, normalIds, diseaseIds)) 
                    list.add(r);
                else
                    diseaseComps.add(r);
            }
        }
        // Now check nodes
        for (Iterator<?> it = comps.iterator(); it.hasNext();) {
            Renderable r = (Renderable) it.next();
            if (!(r instanceof HyperEdge)) {
                if (isNormalObject(r, normalIds, diseaseIds)) 
                    list.add(r);
                else
                    diseaseComps.add(r);
            }
        }
        drawComponents(g, 
                       list,
                       clip);
        if (isForNormal) {
            // Replace the original components with new set of Renderable objects for further processing
            RenderablePathway pathway = (RenderablePathway) displayedObject;
            List<Renderable> components = pathway.getComponents();
            components.clear();
            components.addAll(list);
            components.addAll(compartments); // Don't forget them
            return;
        }
        // Draw a transparent background: light grey
        Color color = new Color(204, 204, 204, 175);
        g2.setPaint(color);
//        Dimension size = getPreferredSize();
        // Preferred size has been scaled. Need to scale it back
        g2.fillRect(0, 0, (int)(size.width / scaleX + 1.0d), (int)(size.height / scaleY + 1.0d));
        // Draw disease related objects
        drawComponents(g, 
                       diseaseComps,
                       clip);
    }
    
    private boolean isNormalObject(Renderable r, Set<Long> normalIds, Set<Long> diseaseIds) {
        if (r instanceof HyperEdge) {
            if (r instanceof FlowLine)
                return isNormalFlowLine((FlowLine)r, 
                                        normalIds,
                                        diseaseIds);
            // Most likely the following statements should not be reached. But place
            // here just in case.
            if (r.getReactomeId() == null)
                return true;
            if (normalIds.contains(r.getReactomeId()))
                return true;
        }
        else if (r instanceof Node) {
            if (r instanceof ProcessNode)
                return true; // Should be used only for normal pathway.
            if (r.getReactomeId() == null)
                return true;
            // Check if it is linked to another disease entities
            Node node = (Node) r;
            List<HyperEdge> edges = node.getConnectedReactions();
            for (HyperEdge edge :edges) {
                if (edge instanceof FlowLine)
                    return isNormalFlowLine((FlowLine)edge, 
                                            normalIds, 
                                            diseaseIds);
                if (edge.getReactomeId() == null) { // Other stuff. This may not be correct!
                    return true;
                }
                if (normalIds.contains(edge.getReactomeId()))
                    return true;
            }
        }
        return false;
    }
    
    /**
     * If a FlowLine connecting a normal object, it should be a normal object. Otherwise, it should
     * be a disease object.
     * @param flowLine
     * @param normalIds
     * @param diseaseIds
     * @return
     */
    private boolean isNormalFlowLine(FlowLine flowLine, 
                                     Set<Long> normalIds,
                                     Set<Long> diseaseIds) {
        Node input = flowLine.getInputNode(0);
        Node output = flowLine.getOutputNode(0);
        if (input != null && normalIds.contains(input.getReactomeId()))
            return true;
        if (output != null && normalIds.contains(output.getReactomeId()))
            return true;
        return false;
    }

    private void drawComponents(Graphics g, 
                                List<Renderable> comps,
                                Rectangle clip) {
        // Draw complexes now
        drawComplexes(comps, clip, g);
        List<HyperEdge> edges = new ArrayList<HyperEdge>();
        for (Renderable obj : comps) {
            if (obj instanceof RenderableCompartment ||
                obj instanceof RenderableComplex ||
                obj instanceof RenderablePathway)
                continue; // Escape it. It should be drawn earlier.
            if (obj instanceof Node) {
                Node node = (Node) obj;
                if (getHidePrivateNote() && (node instanceof Note) && ((Note)node).isPrivate())
                    continue;
                node.validateBounds(g);
                if (clip.intersects(node.getBounds()))
                    node.render(g);
            }
            else if (obj instanceof HyperEdge) {
                edges.add((HyperEdge)obj);
            }
        }
        // Draw HyperEdges
        for (HyperEdge reaction : edges) {
            reaction.validateConnectInfo();
            if (clip.intersects(reaction.getBounds()))
                reaction.render(g);
        }
    }
    
    private void splitObjects(Set<Long> normalIds, Set<Long> diseaseIds)  {
        RenderablePathway diagram = (RenderablePathway) getRenderable();
        if (diagram.getReactomeDiagramId() == null)
            return;
        try {
            GKInstance pdInst = adaptor.fetchInstance(diagram.getReactomeDiagramId());
            // Get the disease pathway instance
            List<?> pathways = pdInst.getAttributeValuesList(ReactomeJavaConstants.representedPathway);
            GKInstance diseasePathway = null;
            for (Object obj : pathways) {
                GKInstance pathway = (GKInstance) obj;
                GKInstance disease = (GKInstance) pathway.getAttributeValue(ReactomeJavaConstants.disease);
                if (disease != null) {
                    diseasePathway = pathway;
                    break;
                }
            }
            if (diseasePathway == null)
                return;
            // There should be a normal pathway contained by a disease pathway
            List<?> subPathways = diseasePathway.getAttributeValuesList(ReactomeJavaConstants.hasEvent);
            GKInstance normalPathway = null;
            for (Object obj : subPathways) {
                GKInstance subPathway = (GKInstance) obj;
                GKInstance disease = (GKInstance) subPathway.getAttributeValue(ReactomeJavaConstants.disease);
                if (disease == null && pathways.contains(subPathway)) {
                    normalPathway = subPathway;
                    break;
                }
            }
            // Get all disease related objects
            Set<GKInstance> normalEvents = InstanceUtilities.grepPathwayEventComponents(normalPathway);
            for (GKInstance inst : normalEvents)
                normalIds.add(inst.getDBID());
            Set<GKInstance> normalEntities = InstanceUtilities.grepPathwayParticipants(normalPathway);
            for (GKInstance inst : normalEntities)
                normalIds.add(inst.getDBID());
            Set<GKInstance> allEvents = InstanceUtilities.grepPathwayEventComponents(diseasePathway);
            allEvents.removeAll(normalEvents);
            for (GKInstance inst : allEvents)
                diseaseIds.add(inst.getDBID());
            Set<GKInstance> allEntities = InstanceUtilities.grepPathwayParticipants(diseasePathway);
            allEntities.removeAll(normalEntities);
            for (GKInstance inst : allEntities)
                diseaseIds.add(inst.getDBID());
        }
        catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
